pub fn sql_for_table<T: crate::model::Entity>() -> (String, String) {
    let types = T::column_types();

    let mut columns = vec!["id integer primary key".to_owned()];

    for i in 1..T::column_count() {
        let col = <T::Column as std::convert::TryFrom<usize>>::try_from(i).unwrap();

        let fk = T::foreign_keys()
            .iter()
            .filter(|x| x.local_column() == &col)
            .take(1);

        let fk = fk.map(|x| {
            format!(
                " references \"{}\"(\"{}\") ON DELETE CASCADE",
                x.foreign_table_name(),
                x.foreign_column_name()
            )
        });

        columns.push(format!(
            "\"{}\" {}{}",
            T::name(col),
            types[i],
            fk.last().unwrap_or_else(|| "".to_string())
        ));
    }

    (
        format!(
            "DROP TABLE IF EXISTS \"{}\"",
            <T as crate::model::Entity>::table_name()
        ),
        format!(
            "CREATE TABLE IF NOT EXISTS \"{}\" ({})",
            <T as crate::model::Entity>::table_name(),
            columns.join(",")
        ),
    )
}

pub fn sql_for_index<I: super::Index>() -> (String, String) {
    use super::Entity;
    (
        format!("DROP INDEX IF EXISTS \"{}\"", I::index_name()),
        format!(
            "CREATE {}INDEX \"{}\" ON \"{}\" ({})",
            if I::unique() { "UNIQUE " } else { "" },
            I::index_name(),
            I::IndexedEntity::table_name(),
            I::columns()
                .iter()
                .map(|x| format!("\"{}\"", I::IndexedEntity::name(*x)))
                .collect::<Vec<_>>()
                .join(",")
        ),
    )
}

#[cfg(test)]
mod test {
    #[derive(serde::Serialize, serde::Deserialize, crate::Entity)]
    #[microrm_internal]
    pub struct Empty {}

    #[derive(serde::Serialize, serde::Deserialize, crate::Entity)]
    #[microrm_internal]
    pub struct Single {
        e: i32,
    }

    #[derive(serde::Serialize, serde::Deserialize, crate::Entity)]
    #[microrm_internal]
    pub struct Reference {
        e: SingleID,
    }

    #[test]
    fn example_sql_for() {
        assert_eq!(
            super::sql_for_table::<Empty>(),
            (
                r#"DROP TABLE IF EXISTS "empty""#.to_owned(),
                r#"CREATE TABLE IF NOT EXISTS "empty" (id integer primary key)"#.to_owned()
            )
        );
        assert_eq!(
            super::sql_for_table::<Single>(),
            (
                r#"DROP TABLE IF EXISTS "single""#.to_owned(),
                r#"CREATE TABLE IF NOT EXISTS "single" (id integer primary key,"e" integer)"#
                    .to_owned()
            )
        );

        assert_eq!(
            super::sql_for_table::<Reference>(),
            (
                r#"DROP TABLE IF EXISTS "reference""#.to_owned(),
                r#"CREATE TABLE IF NOT EXISTS "reference" (id integer primary key,"e" integer)"#
                    .to_owned()
            )
        );
    }

    #[derive(serde::Serialize, serde::Deserialize, crate::Modelable)]
    #[microrm_internal]
    pub struct Unit(u8);
    #[derive(serde::Serialize, serde::Deserialize, crate::Entity)]
    #[microrm_internal]
    pub struct UnitNewtype {
        newtype: Unit,
    }

    #[test]
    fn unit_newtype_struct() {
        assert_eq!(
            super::sql_for_table::<UnitNewtype>(),
            (
                r#"DROP TABLE IF EXISTS "unit_newtype""#.to_owned(),
                r#"CREATE TABLE IF NOT EXISTS "unit_newtype" (id integer primary key,"newtype" integer)"#
                    .to_owned()
            )
        );
    }

    #[derive(serde::Serialize, serde::Deserialize, crate::Modelable)]
    #[microrm_internal]
    pub struct NonUnit(u8, u8);
    #[derive(serde::Serialize, serde::Deserialize, crate::Entity)]
    #[microrm_internal]
    pub struct NonUnitNewtype {
        newtype: NonUnit,
    }

    #[test]
    #[should_panic]
    fn nonunit_newtype_struct() {
        super::sql_for_table::<NonUnitNewtype>();
    }

    #[derive(serde::Serialize, serde::Deserialize, crate::Entity)]
    #[microrm_internal]
    pub struct Child {
        #[microrm_foreign]
        parent_id: SingleID,
    }

    #[test]
    fn test_foreign_key() {
        assert_eq!(
            super::sql_for_table::<Child>(),
            (
                r#"DROP TABLE IF EXISTS "child""#.to_owned(),
                r#"CREATE TABLE IF NOT EXISTS "child" (id integer primary key,"parent_id" integer references "single"("id") ON DELETE CASCADE)"#.to_owned()
            )
        );
    }

    #[derive(serde::Serialize, serde::Deserialize, crate::Entity)]
    #[microrm_internal]
    pub struct KeyValue {
        key: String,
        value: String,
    }

    microrm_macros::make_index_internal!(ValueIndex, KeyValueColumns::Value);
    microrm_macros::make_index_internal!(!UniqueValueIndex, KeyValueColumns::Value);

    #[test]
    fn test_indexes() {
        assert_eq!(
            super::sql_for_index::<ValueIndex>(),
            (
                r#"DROP INDEX IF EXISTS "value_index""#.to_owned(),
                r#"CREATE INDEX "value_index" ON "key_value" ("value")"#.to_owned()
            )
        );
        assert_eq!(
            super::sql_for_index::<UniqueValueIndex>(),
            (
                r#"DROP INDEX IF EXISTS "unique_value_index""#.to_owned(),
                r#"CREATE UNIQUE INDEX "unique_value_index" ON "key_value" ("value")"#.to_owned()
            )
        )
    }

    #[derive(serde::Serialize, serde::Deserialize, crate::Entity)]
    #[microrm_internal]
    pub struct VecTest {
        e: u64,
        test: Vec<usize>
    }

    #[test]
    fn test_vec() {
        assert_eq!(
            super::sql_for_table::<VecTest>().1,
            r#"CREATE TABLE IF NOT EXISTS "vec_test" (id integer primary key,"e" integer,"test" blob)"#.to_owned()
        );
    }

    #[derive(crate::Modelable, serde::Deserialize, serde::Serialize)]
    #[microrm_internal]
    pub enum TestEnum {
        A,
        B,
        C
    }

    #[derive(serde::Serialize, serde::Deserialize, crate::Entity)]
    #[microrm_internal]
    pub struct EnumContainer {
        before: usize,
        e: TestEnum,
        after: usize
    }

    #[test]
    fn test_enum() {
        assert_eq!(
            super::sql_for_table::<EnumContainer>().1,
            r#"CREATE TABLE IF NOT EXISTS "enum_container" (id integer primary key,"before" integer,"e" blob,"after" integer)"#.to_owned()
        )
    }
}
